from enum import Enum, IntEnum


class EngineChoice(str, Enum):
    LOCAL = 'local'
    PREFECT = 'prefect'


class RunState(IntEnum):
    INITIALIZED = 1
    RUNNING = 2
    FINISHED = 3


# TODO: Add 'apply' state
# TODO: Add 'failed' state and error management
# TODO: Consider the need for a 'waiting' state


class RunStateLogMessages(str, Enum):
    INITIALIZED = 'Initialized "{}"'
    RUNNING = 'Started running "{}"...'
    FINISHED = 'Finished running "{}"!'


UNIFAIR_LOG_FORMAT_STR: str = '%(levelname)s (%(name)s) - %(message)s'



__version__ = '0.1.0'

import os
import sys

from omnipy.config.runtime import Runtime

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

runtime = Runtime() if 'pytest' not in sys.modules else None



from collections import defaultdict
import inspect
from typing import DefaultDict, Dict, List, Protocol, Type

from omnipy.util.helpers import create_merged_dict


class IsMixin(Protocol):
    def __init__(self, **kwargs: object) -> None:
        ...


class DynamicMixinAcceptorFactory(type):
    def __new__(mcs, name, bases, namespace):
        cls = type.__new__(mcs, name, bases, namespace)

        if name != 'DynamicMixinAcceptor':
            cls._init_cls_from_metaclass(bases)

        return cls


class DynamicMixinAcceptor(metaclass=DynamicMixinAcceptorFactory):
    # Constants
    WITH_MIXINS_CLS_PREFIX = 'WithMixins'

    # Declarations needed by mypy
    _orig_class: Type
    _orig_init_signature: inspect.Signature
    _mixin_classes: List[Type]
    _init_params_per_mixin_cls: DefaultDict[str, DefaultDict[str, inspect.Parameter]]

    @classmethod
    def _init_cls_from_metaclass(cls, bases):
        if any(base.__name__ == 'DynamicMixinAcceptor' for base in bases):
            if '__init__' not in cls.__dict__:
                raise TypeError('Mixin acceptor class is required to define a __init__() method.')

        cls._orig_init_signature = inspect.signature(cls.__init__)
        cls._mixin_classes = []
        cls._init_params_per_mixin_cls = defaultdict(defaultdict)

        for base in bases:
            if hasattr(base, '_mixin_classes'):
                for mixin_class in base._mixin_classes:
                    if mixin_class.__name__ not in cls._init_params_per_mixin_cls:
                        cls._mixin_classes.append(mixin_class)
                        cls._init_params_per_mixin_cls[mixin_class.__name__] = \
                            base._init_params_per_mixin_cls[mixin_class.__name__]

    @classmethod
    def _get_mixin_init_kwarg_params(cls) -> Dict[str, inspect.Parameter]:
        return {
            key: param for param_dict in cls._init_params_per_mixin_cls.values() for key,
            param in param_dict.items()
        }

    @property
    def _mixin_init_kwarg_params(self) -> Dict[str, inspect.Parameter]:
        return self._get_mixin_init_kwarg_params()

    @classmethod
    def accept_mixin(cls, mixin_cls: Type) -> None:
        cls._mixin_classes.append(mixin_cls)

        if '__init__' in mixin_cls.__dict__:
            cls._store_init_signature_params_for_mixin(mixin_cls)
            cls._update_cls_init_signature_with_kwargs_for_all_mixins()

    @classmethod
    def _store_init_signature_params_for_mixin(cls, mixin_cls):
        for key, param in inspect.signature(mixin_cls.__init__).parameters.items():
            if key != 'self':
                if param.kind != param.KEYWORD_ONLY:
                    raise AttributeError(
                        'All params in the signature of the __init__() method in a dynamic '
                        'mixin class must be keyword-only (except for the "self" param)')
                cls._init_params_per_mixin_cls[mixin_cls.__name__][key] = param

    @classmethod
    def _update_cls_init_signature_with_kwargs_for_all_mixins(cls):
        orig_init_param_dict = cls._orig_init_signature.parameters

        init_params = list(orig_init_param_dict.values())
        if init_params[-1].kind != inspect.Parameter.VAR_KEYWORD:
            raise AttributeError('Mixin acceptor class is required to define a catch-all "kwargs"'
                                 ' parameter as the last parameter of its __init__() method.')

        var_keyword_param = init_params[-1]
        init_params = init_params[:-1]
        only_mixin_params = [
            val for (key, val) in cls._get_mixin_init_kwarg_params().items()
            if key not in orig_init_param_dict
        ]

        updated_params = init_params + only_mixin_params + [var_keyword_param]
        updated_init_signature = cls._orig_init_signature.replace(parameters=updated_params)

        cls.__init__.__signature__ = updated_init_signature

    def __new__(cls, *args, **kwargs):
        if not cls.__name__.endswith(cls.WITH_MIXINS_CLS_PREFIX):
            cls_with_mixins = cls._create_subcls_inheriting_from_mixins_and_orig_cls()
            return object.__new__(cls_with_mixins)
        else:
            return object.__new__(cls)

    @classmethod
    def _create_subcls_inheriting_from_mixins_and_orig_cls(cls):
        def _initialize_mixins(self, kwargs):
            for mixin_cls in self._mixin_classes:
                mixin_kwargs = {}

                for key, param in self._init_params_per_mixin_cls[mixin_cls.__name__].items():
                    if key in kwargs:
                        mixin_kwargs[key] = kwargs[key]

                mixin_cls.__init__(self, **mixin_kwargs)

        def _check_mixin_kwarg_defaults_and_return_if_not_in_kwargs(self, kwargs):
            orig_init_param_dict = cls._orig_init_signature.parameters

            kwarg_new_param_default = {}
            kwarg_new_param_default_mixin_name = {}

            for mixin_cls in self._mixin_classes:
                for key, param in self._init_params_per_mixin_cls[mixin_cls.__name__].items():
                    if param.default is not param.empty:
                        if key in orig_init_param_dict and \
                                param.default != orig_init_param_dict[key].default:
                            raise AttributeError(
                                f'Default value for keyword argument "{key}" differs between '
                                f'__init__() methods of mixin class "{mixin_cls.__name__}" '
                                f'and original class "{self._orig_class.__name__}": '
                                f'{param.default} != {orig_init_param_dict[key].default}')
                        elif key in kwarg_new_param_default and \
                                param.default != kwarg_new_param_default[key]:
                            raise AttributeError(
                                f'Default value for keyword argument "{key}" differs between '
                                f'__init__() methods of mixin classes "{mixin_cls.__name__}" and'
                                f'"{kwarg_new_param_default_mixin_name[key]}": '
                                f'{param.default} != {kwarg_new_param_default[key]}')
                        else:
                            kwarg_new_param_default[key] = param.default
                            kwarg_new_param_default_mixin_name[key] = mixin_cls.__name__

            return {
                key: param_default for key,
                param_default in kwarg_new_param_default.items() if key not in kwargs
            }

        def __init__(self, *args, **kwargs):
            mixin_kwargs_defaults = \
                _check_mixin_kwarg_defaults_and_return_if_not_in_kwargs(self, kwargs)

            self._orig_class.__init__(self,
                                      *args,
                                      **create_merged_dict(kwargs, mixin_kwargs_defaults))
            _initialize_mixins(self, kwargs)

        cls_with_mixins = DynamicMixinAcceptorFactory(
            f'{cls.__name__}{cls.WITH_MIXINS_CLS_PREFIX}',
            tuple(list(reversed(cls._mixin_classes)) + [cls]),
            dict(__init__=__init__),
        )

        cls_with_mixins._orig_class = cls
        cls_with_mixins._orig_init_signature = inspect.signature(cls.__init__)

        cls_with_mixins._update_cls_init_signature_with_kwargs_for_all_mixins()  # noqa

        return cls_with_mixins



from __future__ import annotations

from typing import cast, Type

from omnipy.compute.job import CallableDecoratingJobTemplateMixin
from omnipy.compute.private.job import FuncJob, FuncJobBase, FuncJobTemplate
from omnipy.compute.types import FuncJobTemplateCallable
from omnipy.engine.protocols import IsTask
from omnipy.util.callable_decorator_cls import callable_decorator_cls


class TaskBase(FuncJobBase):
    ...


def task_template_callable_decorator_cls(
        cls: Type['TaskTemplate']) -> FuncJobTemplateCallable['TaskTemplate']:
    return cast(FuncJobTemplateCallable['TaskTemplate'], callable_decorator_cls(cls))


@task_template_callable_decorator_cls
class TaskTemplate(CallableDecoratingJobTemplateMixin['TaskTemplate'],
                   TaskBase,
                   FuncJobTemplate['Task']):
    @classmethod
    def _get_job_subcls_for_apply(cls) -> Type[Task]:
        return Task

    @classmethod
    def _apply_engine_decorator(cls, task: IsTask) -> IsTask:
        if cls.engine is not None:
            return cls.engine.task_decorator(task)
        else:
            return task


class Task(TaskBase, FuncJob[TaskBase, TaskTemplate]):
    @classmethod
    def _get_job_base_subcls_for_init(cls) -> Type[TaskBase]:
        return TaskBase

    @classmethod
    def _get_job_template_subcls_for_revise(cls) -> Type[TaskTemplate]:
        return TaskTemplate


# TODO: Would we need the possibility to refine task templates by adding new task parameters?



from dataclasses import dataclass
from enum import Enum


class GlobalSerializeOutputsOptions(Enum):
    OFF = 0
    WRITE_FLOW_OUTPUTS = 1
    WRITE_FLOW_AND_TASK_OUTPUTS = 2


class GlobalResumePreviousRunsOptions(Enum):
    OFF = 0
    AUTO_READ_IGNORE_PARAMS = 1


@dataclass
class JobConfig:
    serialize_outputs: GlobalSerializeOutputsOptions = \
        GlobalSerializeOutputsOptions.WRITE_FLOW_AND_TASK_OUTPUTS
    resume_previous_runs: GlobalResumePreviousRunsOptions = \
        GlobalResumePreviousRunsOptions.OFF



from inspect import Parameter, signature
from typing import Optional, Type

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model


class IterateFuncJobBaseMixin:
    # Requires FuncSignatureJobBaseMixin

    def __init__(self, *, iterate_over_data_files: bool = False):
        self._iterate_over_data_files = iterate_over_data_files

        if not isinstance(self.iterate_over_data_files, bool):
            raise TypeError(
                'Value of "iterate_over_data_files" parameter must be bool (True/False), '
                f'not "{iterate_over_data_files}"')

        if iterate_over_data_files:
            inner_func = self._job_func
            if inner_func.__name__ != '_omnipy_iterate_func':
                func_signature = signature(inner_func)
                params = list(func_signature.parameters.values())

                if len(params) == 0 or \
                        params[0].default != Parameter.empty or \
                        params[0].kind not in [Parameter.POSITIONAL_ONLY,
                                               Parameter.POSITIONAL_OR_KEYWORD]:
                    raise AttributeError(
                        'Parameter "iterate_over_data_files" is set to True,'
                        'but the job function has no arguments without default values. '
                        'The first argument will be replaced with a corresponding '
                        'Dataset arg to be iterated over')

                data_param = params[0]
                rest_params = params[1:]

                def create_dataset_cls(data_file_type: Type) -> Dataset[Model]:
                    return Dataset[data_file_type] if issubclass(data_file_type, Model) \
                        else Dataset[Model[data_file_type]]

                dataset_cls = create_dataset_cls(data_param.annotation)
                out_dataset_cls = create_dataset_cls(func_signature.return_annotation)
                dataset_param = data_param.replace(name='dataset', annotation=dataset_cls)
                new_signature = func_signature.replace(
                    parameters=[dataset_param] + rest_params,
                    return_annotation=out_dataset_cls,
                )

                def _omnipy_iterate_func(dataset: dataset_cls, *args: object, **kwargs: object):
                    out_dataset = out_dataset_cls()
                    for title, data_file in dataset.items():
                        out_dataset[title] = inner_func(data_file, *args, **kwargs)
                    return out_dataset

                self._job_func = _omnipy_iterate_func
                self._job_func.__signature__ = new_signature
                self._func_signature = new_signature
                self.__signature__ = new_signature

    @property
    def iterate_over_data_files(self) -> Optional[str]:
        return self._iterate_over_data_files



from typing import Any, Dict, IO

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model
from omnipy.data.serializer import create_dataset_from_tarfile, create_tarfile_from_dataset


class RawDatasetToTarFileSerializer:
    @staticmethod
    def serialize(dataset: Dataset[Model[str]]) -> bytes:
        def raw_encode_func(contents: str) -> bytes:
            return contents.encode('utf8')

        return create_tarfile_from_dataset(
            dataset, file_suffix='raw', data_encode_func=raw_encode_func)

    @staticmethod
    def deserialize(tarfile_bytes: bytes) -> Dataset[Model[str]]:
        dataset = Dataset[Model[str]]()

        def raw_decode_func(file_stream: IO[bytes]) -> str:
            return file_stream.read().decode('utf8')

        def python_dictify_object(obj_type: str, obj_val: Any) -> Dict:
            return {obj_type: obj_val}

        create_dataset_from_tarfile(
            dataset,
            tarfile_bytes,
            file_suffix='raw',
            data_decode_func=raw_decode_func,
            dictify_object_func=python_dictify_object,
            import_method='from_data')  # noqa

        return dataset



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from functools import update_wrapper
from types import MappingProxyType
from typing import Any, Dict, Generic, Hashable, Optional, Tuple, Type, Union

from omnipy.compute.mixins.name import NameJobBaseMixin, NameJobMixin
from omnipy.compute.types import JobBaseT, JobT, JobTemplateT
from omnipy.config.job import JobConfig
from omnipy.engine.protocols import IsJob, IsJobConfig, IsJobCreator, IsTaskRunnerEngine
from omnipy.util.helpers import create_merged_dict
from omnipy.util.mixin import DynamicMixinAcceptor, DynamicMixinAcceptorFactory


class JobCreator:
    def __init__(self) -> None:
        self._engine: Optional[IsTaskRunnerEngine] = None
        self._config: IsJobConfig = JobConfig()
        self._nested_context_level = 0

    def set_engine(self, engine: IsTaskRunnerEngine) -> None:
        self._engine = engine

    def set_config(self, config: IsJobConfig) -> None:
        self._config = config

    def __enter__(self):
        self._nested_context_level += 1

    def __exit__(self, exc_type, exc_value, traceback):
        self._nested_context_level -= 1

    @property
    def engine(self) -> Optional[IsTaskRunnerEngine]:
        return self._engine

    @property
    def config(self) -> IsJobConfig:
        return self._config

    @property
    def nested_context_level(self) -> int:
        return self._nested_context_level


class JobBaseMeta(ABCMeta):
    _job_creator: IsJobCreator = JobCreator()

    @property
    def job_creator(self) -> IsJobCreator:
        return self._job_creator


class JobTemplateMeta(JobBaseMeta):
    @property
    def engine(self) -> Optional[IsTaskRunnerEngine]:
        return self.job_creator.engine

    @property
    def nested_context_level(self) -> int:
        return self.job_creator.nested_context_level


class JobBaseAndMixinAcceptorMeta(JobBaseMeta, DynamicMixinAcceptorFactory):
    ...


class JobTemplateAndMixinAcceptorMeta(JobTemplateMeta, JobBaseAndMixinAcceptorMeta):
    ...


class JobBase(DynamicMixinAcceptor, metaclass=JobBaseAndMixinAcceptorMeta):
    def __init__(self, *args: object, name: Optional[str] = None, **kwargs: object):
        super().__init__(*args, **kwargs)

        if not isinstance(self, JobTemplate) and not isinstance(self, Job):
            raise RuntimeError('JobBase and subclasses not inheriting from JobTemplate '
                               'or Job are not directly instantiatable')

    @abstractmethod
    def _get_init_arg_values(self) -> Union[Tuple[()], Tuple[Any, ...]]:
        ...

    def _get_init_args(self) -> Union[Tuple[()], Tuple[Any, ...]]:
        return self._get_init_arg_values()

    @abstractmethod
    def _get_init_kwarg_public_property_keys(self) -> Tuple[str, ...]:
        ...

    def _get_init_kwargs(self) -> Dict[str, Any]:
        kwarg_keys = list(self._mixin_init_kwarg_params.keys())
        kwarg_keys += list(self._get_init_kwarg_public_property_keys())
        for key in kwarg_keys:
            attribute = getattr(self.__class__, key)

            if not isinstance(attribute, property):
                raise TypeError(f'{self.__class__.__name__} attribute "{key}" is not a property')

            if attribute.fset is not None:
                raise TypeError(f'{self.__class__.__name__} attribute "{key}" is not a property')

            value = getattr(self, key)
            if not isinstance(value, Hashable) and not isinstance(value, MappingProxyType):
                raise TypeError(f'Value of {self.__class__.__name__} attribute "{key}" is mutable')

        return {key: getattr(self, key) for key in kwarg_keys}

    def __eq__(self, other: object):
        if not isinstance(other, JobBase):
            return NotImplemented
        return self._get_init_args() == other._get_init_args() \
            and self._get_init_kwargs() == other._get_init_kwargs()

    @property
    def in_flow_context(self) -> bool:
        return self.__class__.job_creator.nested_context_level > 0


JobBase.accept_mixin(NameJobBaseMixin)


class JobTemplate(JobBase, Generic[JobT], metaclass=JobTemplateAndMixinAcceptorMeta):
    @classmethod
    @abstractmethod
    def _get_job_subcls_for_apply(cls) -> Type[JobT]:
        return Job

    @property
    def engine(self) -> Optional[IsTaskRunnerEngine]:
        return self.__class__.engine

    @abstractmethod
    def _apply_engine_decorator(self, job: IsJob) -> IsJob:
        ...

    @classmethod
    def create(cls, *init_args: object, **init_kwargs: object) -> JobTemplateT:
        return cls(*init_args, **init_kwargs)

    def run(self, *args: object, **kwargs: object):
        return self.apply()(*args, **kwargs)

    def apply(self) -> JobT:
        job_cls = self._get_job_subcls_for_apply()
        job = job_cls.create(*self._get_init_args(), **self._get_init_kwargs())
        update_wrapper(job, self, updated=[])
        return self._apply_engine_decorator(job)

    def refine(self, *args: object, update: bool = True, **kwargs: object) -> JobTemplateT:
        if update:
            for key, cur_val in self._get_init_kwargs().items():
                if key in kwargs:
                    try:
                        new_val = create_merged_dict(cur_val, kwargs[key])
                    except (TypeError, ValueError):
                        new_val = kwargs[key]
                else:
                    new_val = cur_val
                kwargs[key] = new_val

        return self.create(*(args if args else self._get_init_args()), **kwargs)

    def __call__(self, *args, **kwargs):
        if self.in_flow_context:
            job = self.apply()
            return job(*args, **kwargs)
        raise TypeError(f"'{self.__class__.__name__}' object is not callable")


class CallableDecoratingJobTemplateMixin(Generic[JobTemplateT]):
    @classmethod
    def create(cls: Type[JobTemplateT], *init_args: object, **init_kwargs: object) -> JobTemplateT:
        return cls(*(init_args[1:]), **init_kwargs)(init_args[0])


class Job(JobBase, DynamicMixinAcceptor, Generic[JobBaseT, JobTemplateT]):
    @property
    def flow_context(self) -> IsJobCreator:
        return self.__class__.job_creator

    @classmethod
    @abstractmethod
    def _get_job_base_subcls_for_init(cls) -> Type[JobBaseT]:
        return JobBase

    @classmethod
    @abstractmethod
    def _get_job_template_subcls_for_revise(cls) -> Type[JobTemplateT]:
        return JobTemplate

    def __new__(cls, *args: Any, **kwargs: Any):
        super().__new__(cls, *args, **kwargs)
        raise RuntimeError('Job should only be instantiated using the "apply()" method of '
                           'an instance of JobTemplate (or one of its subclasses)')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def create(cls, *init_args: object, **init_kwargs: object) -> Job[JobBaseT, JobTemplateT]:
        job_base_cls = cls._get_job_base_subcls_for_init()
        if job_base_cls.__new__ is object.__new__:
            job_obj = job_base_cls.__new__(cls)
        else:
            job_obj = job_base_cls.__new__(cls, *init_args, **init_kwargs)
        job_obj.__init__(*init_args, **init_kwargs)
        return job_obj

    def revise(self) -> JobTemplateT:
        job_template_cls = self._get_job_template_subcls_for_revise()
        job_template = job_template_cls.create(*self._get_init_args(), **self._get_init_kwargs())
        update_wrapper(job_template, self, updated=[])
        return job_template

    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        ...

    @abstractmethod
    def _call_func(self, *args: Any, **kwargs: Any) -> Any:
        ...


Job.accept_mixin(NameJobMixin)

# TODO: Change JobBase and friends into Generics such as one can annotated with
#       e.g. 'TaskTemplate[[int], int]' instead of just 'TaskTemplate'



from io import StringIO
import os
from types import NoneType
from typing import Callable, List, Union

from omnipy.compute.task import TaskTemplate
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model


@TaskTemplate
def modify_datafile_contents(
    dataset: Dataset[Model[str]],
    modify_contents_func: Callable[[str], str],
    **kwargs: object,
) -> Dataset[Model[str]]:
    out_dataset = Dataset[Model[str]]()
    for name, datafile in dataset.items():
        out_dataset[name] = modify_contents_func(datafile, **kwargs)
    return out_dataset


@TaskTemplate
def modify_each_line(
    dataset: Dataset[Model[str]],
    modify_line_func: Callable[[int, str], Union[NoneType, str]],
    **kwargs: object,
) -> Dataset[Model[str]]:
    out_dataset = Dataset[Model[str]]()
    for name, datafile in dataset.items():
        output_data = StringIO()
        for i, line in enumerate(StringIO(datafile)):
            modified_line = modify_line_func(i, line, **kwargs)
            if modified_line is not None:
                output_data.write(modified_line)
        out_dataset[name] = output_data.getvalue()
    return out_dataset


@TaskTemplate
def modify_all_lines(
    dataset: Dataset[Model[str]],
    modify_all_lines_func: Callable[[int, List[str]], List[str]],
    **kwargs: object,
) -> Dataset[Model[str]]:
    out_dataset = Dataset[Model[str]]()
    for name, datafile in dataset.items():
        all_lines = [line.strip() for line in StringIO(datafile)]
        modified_lines = modify_all_lines_func(all_lines, **kwargs)
        out_dataset[name] = os.linesep.join(modified_lines)
    return out_dataset



from abc import ABC, abstractmethod
from io import BytesIO
import tarfile
from tarfile import TarInfo
from typing import Any, Callable, Dict, IO, Union

from omnipy.data.dataset import Dataset


class Serializer(ABC):
    @staticmethod
    @abstractmethod
    def serialize(dataset: Dataset):
        pass

    @staticmethod
    @abstractmethod
    def deserialize(serialized) -> Dataset:
        pass


def create_tarfile_from_dataset(dataset: Dataset,
                                file_suffix: str,
                                data_encode_func: Callable[[Any], Union[bytes, memoryview]]):
    bytes_io = BytesIO()
    with tarfile.open(fileobj=bytes_io, mode='w:gz') as tarfile_stream:
        for obj_type, data_obj in dataset.items():
            json_data_bytestream = BytesIO(data_encode_func(data_obj))
            json_data_bytestream.seek(0)
            tarinfo = TarInfo(name=f'{obj_type}.{file_suffix}')
            tarinfo.size = len(json_data_bytestream.getbuffer())
            tarfile_stream.addfile(tarinfo, json_data_bytestream)
    return bytes_io.getbuffer().tobytes()


def create_dataset_from_tarfile(dataset: Dataset,
                                tarfile_bytes: bytes,
                                file_suffix: str,
                                data_decode_func: Callable[[IO[bytes]], Any],
                                dictify_object_func: Callable[[str, Any], Union[Dict, str]],
                                import_method='from_data'):
    with tarfile.open(fileobj=BytesIO(tarfile_bytes), mode='r:gz') as tarfile_stream:
        for filename in tarfile_stream.getnames():
            obj_type_file = tarfile_stream.extractfile(filename)
            assert filename.endswith(f'.{file_suffix}')
            obj_type = '.'.join(filename.split('.')[:-1])
            getattr(dataset, import_method)(
                dictify_object_func(obj_type, data_decode_func(obj_type_file)))


class CsvSerializer:
    pass


class PythonSerializer:
    pass


from typing import cast, Type

from omnipy.compute.job import CallableDecoratingJobTemplateMixin
from omnipy.compute.private.flow import (Flow,
                                         FlowBase,
                                         FlowTemplate,
                                         TaskTemplatesFlow,
                                         TaskTemplatesFlowBase,
                                         TaskTemplatesFlowTemplate)
from omnipy.compute.types import FuncJobTemplateCallable, TaskTemplatesFlowTemplateCallable
from omnipy.engine.protocols import (IsDagFlow,
                                     IsDagFlowRunnerEngine,
                                     IsFuncFlow,
                                     IsFuncFlowRunnerEngine,
                                     IsLinearFlow,
                                     IsLinearFlowRunnerEngine)
from omnipy.util.callable_decorator_cls import callable_decorator_cls


class LinearFlowBase(TaskTemplatesFlowBase):
    ...


def linear_flow_template_callable_decorator_cls(
        cls: Type['LinearFlowTemplate']) -> TaskTemplatesFlowTemplateCallable['LinearFlowTemplate']:
    return cast(TaskTemplatesFlowTemplateCallable['LinearFlowTemplate'],
                callable_decorator_cls(cls))


@linear_flow_template_callable_decorator_cls
class LinearFlowTemplate(CallableDecoratingJobTemplateMixin['LinearFlowTemplate'],
                         LinearFlowBase,
                         TaskTemplatesFlowTemplate['LinearFlow']):
    @classmethod
    def _get_job_subcls_for_apply(cls) -> Type['LinearFlow']:
        return LinearFlow

    def _apply_engine_decorator(self, flow: IsLinearFlow) -> IsLinearFlow:
        if self.engine is not None and isinstance(self.engine, IsLinearFlowRunnerEngine):
            return self.engine.linear_flow_decorator(flow)
        else:
            raise RuntimeError(f'Engine "{self.engine}" does not support linear flows')


class LinearFlow(LinearFlowBase, TaskTemplatesFlow[LinearFlowBase, LinearFlowTemplate]):
    @classmethod
    def _get_job_base_subcls_for_init(cls) -> Type[LinearFlowBase]:
        return LinearFlowBase

    @classmethod
    def _get_job_template_subcls_for_revise(cls) -> Type[LinearFlowTemplate]:
        return LinearFlowTemplate


class DagFlowBase(TaskTemplatesFlowBase):
    ...


def dag_flow_template_callable_decorator_cls(
        cls: Type['DagFlowTemplate']) -> TaskTemplatesFlowTemplateCallable['DagFlowTemplate']:
    return cast(TaskTemplatesFlowTemplateCallable['DagFlowTemplate'], callable_decorator_cls(cls))


@dag_flow_template_callable_decorator_cls
class DagFlowTemplate(CallableDecoratingJobTemplateMixin['DagFlowTemplate'],
                      DagFlowBase,
                      TaskTemplatesFlowTemplate['DagFlow']):
    @classmethod
    def _get_job_subcls_for_apply(cls) -> Type['DagFlow']:
        return DagFlow

    def _apply_engine_decorator(self, job: IsDagFlow) -> IsDagFlow:
        if self.engine is not None and isinstance(self.engine, IsDagFlowRunnerEngine):
            return self.engine.dag_flow_decorator(job)
        else:
            raise RuntimeError(f'Engine "{self.engine}" does not support DAG flows')


class DagFlow(DagFlowBase, TaskTemplatesFlow[DagFlowBase, DagFlowTemplate]):
    @classmethod
    def _get_job_base_subcls_for_init(cls) -> Type[DagFlowBase]:
        return DagFlowBase

    @classmethod
    def _get_job_template_subcls_for_revise(cls) -> Type[DagFlowTemplate]:
        return DagFlowTemplate


class FuncFlowBase(FlowBase):
    ...


def func_flow_template_callable_decorator_cls(
        cls: Type['FuncFlowTemplate']) -> FuncJobTemplateCallable['FuncFlowTemplate']:
    return cast(FuncJobTemplateCallable['FuncFlowTemplate'], callable_decorator_cls(cls))


@func_flow_template_callable_decorator_cls
class FuncFlowTemplate(CallableDecoratingJobTemplateMixin['FuncFlowTemplate'],
                       FuncFlowBase,
                       FlowTemplate['FuncFlow']):
    def _apply_engine_decorator(self, flow: IsFuncFlow) -> IsFuncFlow:
        if self.engine is not None and isinstance(self.engine, IsFuncFlowRunnerEngine):
            return self.engine.func_flow_decorator(flow)
        else:
            raise RuntimeError(f'Engine "{self.engine}" does not support function flows')

    @classmethod
    def _get_job_subcls_for_apply(cls) -> Type['FuncFlow']:
        return FuncFlow


class FuncFlow(FuncFlowBase, Flow[FuncFlowBase, FuncFlowTemplate]):
    @classmethod
    def _get_job_base_subcls_for_init(cls) -> Type[FuncFlowBase]:
        return FuncFlowBase

    @classmethod
    def _get_job_template_subcls_for_revise(cls) -> Type[FuncFlowTemplate]:
        return FuncFlowTemplate


# TODO: Recursive replace - *args: Any -> *args: object, *kwargs: Any -> *kwargs: object



from abc import ABC
from typing import Any, Callable, Generic, Mapping, Optional, Tuple, Union

from omnipy.compute.job import JobTemplate
from omnipy.compute.private.job import FuncJob, FuncJobBase, FuncJobTemplate
from omnipy.compute.types import FlowBaseT, FlowT, FlowTemplateT, TaskTemplatesFlowTemplateT
from omnipy.engine.protocols import IsTaskTemplate
from omnipy.util.helpers import remove_none_vals


class FlowBase(FuncJobBase):
    ...


class FlowTemplate(FuncJobTemplate[FlowT], Generic[FlowT], ABC):
    ...


class Flow(FuncJob[FlowBaseT, FlowTemplateT], Generic[FlowBaseT, FlowTemplateT], ABC):
    ...


class TaskTemplatesFlowBase(FlowBase):
    def __init__(
        self,
        job_func: Callable,
        *task_templates: IsTaskTemplate,
        name: Optional[str] = None,
        fixed_params: Optional[Mapping[str, object]] = None,
        param_key_map: Optional[Mapping[str, str]] = None,
        result_key: Optional[str] = None,
        iterate_over_data_files: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            job_func,
            **remove_none_vals(
                name=name,
                fixed_params=fixed_params,
                param_key_map=param_key_map,
                result_key=result_key,
                iterate_over_data_files=iterate_over_data_files,
                **kwargs,
            ))

        self._task_templates: Tuple[IsTaskTemplate, ...] = task_templates

    def _get_init_arg_values(self) -> Union[Tuple[()], Tuple[Any, ...]]:
        return self._job_func, *self._task_templates

    @property
    def task_templates(self) -> Tuple[IsTaskTemplate, ...]:
        return self._task_templates


class TaskTemplatesFlowTemplate(TaskTemplatesFlowBase, FlowTemplate[FlowT], Generic[FlowT], ABC):
    def refine(self,
               *task_templates: IsTaskTemplate,
               update: bool = True,
               name: Optional[str] = None,
               fixed_params: Optional[Mapping[str, object]] = None,
               param_key_map: Optional[Mapping[str, str]] = None,
               result_key: Optional[str] = None,
               iterate_over_data_files: bool = False,
               **kwargs: Any) -> TaskTemplatesFlowTemplateT:

        args = tuple([self._job_func] + list(*task_templates)) if task_templates else ()

        return JobTemplate.refine(
            self,
            *args,
            update=update,
            **remove_none_vals(
                name=name,
                fixed_params=fixed_params,
                param_key_map=param_key_map,
                result_key=result_key,
                iterate_over_data_files=iterate_over_data_files,
                **kwargs,
            ))


class TaskTemplatesFlow(TaskTemplatesFlowBase,
                        Flow[FlowBaseT, FlowTemplateT],
                        Generic[FlowBaseT, FlowTemplateT],
                        ABC):
    ...



from datetime import datetime
import logging
from typing import Any, Callable, Dict, Optional, Protocol, runtime_checkable, Tuple, Type

from omnipy.config.job import GlobalResumePreviousRunsOptions, GlobalSerializeOutputsOptions
from omnipy.engine.constants import EngineChoice, RunState


class IsJobConfig(Protocol):
    serialize_outputs: GlobalSerializeOutputsOptions = \
        GlobalSerializeOutputsOptions.WRITE_FLOW_AND_TASK_OUTPUTS
    resume_previous_runs: GlobalResumePreviousRunsOptions = \
        GlobalResumePreviousRunsOptions.OFF


class IsEngineConfig(Protocol):
    ...


class IsLocalRunnerConfig(IsEngineConfig, Protocol):
    ...


class IsPrefectEngineConfig(IsEngineConfig, Protocol):
    use_cached_results: int = False


class IsRunStateRegistryConfig(Protocol):
    verbose: bool = True

    def __init__(
            self,
            verbose: bool = True,  # noqa
            *args: Any,
            **kwargs: Any) -> None:
        ...


class IsConfigPublisher(Protocol):
    def subscribe(self, config_item: str, callback_fun: Callable[[Any], None]):
        ...

    def unsubscribe_all(self) -> None:
        ...


class IsRuntimeConfig(IsConfigPublisher, Protocol):
    engine: EngineChoice
    local: IsLocalRunnerConfig
    prefect: IsPrefectEngineConfig
    registry: IsRunStateRegistryConfig

    def __init__(
            self,
            engine: EngineChoice = EngineChoice.LOCAL,  # noqas
            local_config: Optional[IsLocalRunnerConfig] = None,  # noqa
            prefect_config: Optional[IsPrefectEngineConfig] = None,  # noqa
            registry_config: Optional[IsRunStateRegistryConfig] = None,  # noqa
            *args: Any,
            **kwargs: Any) -> None:
        ...


class IsRuntimeObjects(IsConfigPublisher, Protocol):
    logger: logging.Logger
    registry: 'IsRunStateRegistry'
    job_creator: 'IsJobCreator'
    local: 'IsEngine'
    prefect: 'IsEngine'

    def __init__(
            self,
            logger: Optional[logging.Logger] = None,  # noqa
            registry: Optional['IsRunStateRegistry'] = None,  # noqa
            job_creator: Optional['IsJobCreator'] = None,  # noqa
            local: Optional['IsEngine'] = None,  # noqa
            prefect: Optional['IsEngine'] = None,  # noqa
            *args: Any,
            **kwargs: Any) -> None:
        ...


class IsRuntime(IsConfigPublisher, Protocol):
    object: IsRuntimeObjects
    config: IsRuntimeConfig

    def __init__(
            self,
            object: Optional[IsRuntimeObjects] = None,  # noqa
            config: Optional[IsRuntimeConfig] = None,  # noqa
            *args: Any,
            **kwargs: Any) -> None:
        ...


class IsJobCreator(Protocol):
    config: IsJobConfig
    engine: Optional['IsTaskRunnerEngine']
    nested_context_level: int

    def __enter__(self):
        ...

    def __exit__(self, exc_type, exc_value, traceback):
        ...

    def set_config(self, config: IsJobConfig) -> None:
        ...

    def set_engine(self, engine: 'IsTaskRunnerEngine') -> None:
        ...


class IsJob(Protocol):
    name: str
    unique_name: str
    flow_context: IsJobCreator
    in_flow_context: bool

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        ...

    def _call_func(self, *args: Any, **kwargs: Any) -> Any:
        ...

    def regenerate_unique_name(self) -> None:
        ...


class IsJobTemplate(Protocol):
    job_creator: IsJobCreator
    in_flow_context: bool

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        pass


class IsFuncJob(IsJob, Protocol):
    def has_coroutine_func(self) -> bool:
        ...


class IsFuncJobTemplate(IsJobTemplate, Protocol):
    def apply(self) -> IsFuncJob:
        ...


class IsTask(IsFuncJob, Protocol):
    def has_coroutine_func(self) -> bool:
        ...


class IsTaskTemplate(IsFuncJobTemplate, Protocol):
    def apply(self) -> IsTask:
        ...


class IsFlow(IsFuncJob, Protocol):
    def has_coroutine_func(self) -> bool:
        ...


class IsFlowTemplate(IsFuncJobTemplate, Protocol):
    def apply(self) -> IsFlow:
        ...


class IsTaskTemplatesFlow(IsFlow, Protocol):
    task_templates: Tuple[IsTaskTemplate]

    def get_call_args(self, *args, **kwargs) -> Dict[str, object]:
        ...


class IsTaskTemplatesFlowTemplate(IsFuncJobTemplate, Protocol):
    ...


class IsLinearFlow(IsTaskTemplatesFlow, Protocol):
    ...


class IsLinearFlowTemplate(IsLinearFlow, IsTaskTemplatesFlowTemplate, Protocol):
    ...


class IsDagFlow(IsTaskTemplatesFlow, Protocol):
    ...


class IsDagFlowTemplate(IsDagFlow, IsTaskTemplatesFlowTemplate, Protocol):
    ...


class IsFuncFlow(IsFlow, Protocol):
    ...


class IsFuncFlowTemplate(IsFuncFlow, IsFlowTemplate, Protocol):
    ...


class IsEngine(Protocol):
    def __init__(self) -> None:
        ...

    @classmethod
    def get_config_cls(cls) -> Type[IsEngineConfig]:
        ...

    def set_config(self, config: IsEngineConfig) -> None:
        ...

    def set_registry(self, registry: Optional['IsRunStateRegistry']) -> None:
        ...


class IsTaskRunnerEngine(IsEngine, Protocol):
    def task_decorator(self, task: IsTask) -> IsTask:
        ...


@runtime_checkable
class IsLinearFlowRunnerEngine(IsEngine, Protocol):
    def linear_flow_decorator(self, linear_flow: IsLinearFlow) -> IsLinearFlow:
        ...


@runtime_checkable
class IsDagFlowRunnerEngine(IsEngine, Protocol):
    def dag_flow_decorator(self, dag_flow: IsDagFlow) -> IsDagFlow:
        ...


@runtime_checkable
class IsFuncFlowRunnerEngine(IsEngine, Protocol):
    def func_flow_decorator(self, dag_flow: IsFuncFlow) -> IsFuncFlow:
        ...


class IsRunStateRegistry(Protocol):
    def __init__(self) -> None:
        ...

    def get_job_state(self, job: IsJob) -> RunState:
        ...

    def get_job_state_datetime(self, job: IsJob, state: RunState) -> datetime:
        ...

    def all_jobs(self, state: Optional[RunState] = None) -> Tuple[IsJob, ...]:  # noqa
        ...

    def set_job_state(self, job: IsJob, state: RunState) -> None:
        ...

    def set_logger(self, logger: Optional[logging.Logger]) -> None:
        ...

    def set_config(self, config: IsRunStateRegistryConfig) -> None:
        ...
